@@ -22,16 +22,23 @@ open Std_utils
 open Common_gettext.Gettext
 open Getopt.OptionName
 
+(* For this file only, create a reference to prog so that the
+ * --program-name option can change the program name in error
+ * messages.
+ *)
+let prog = ref prog
+
 type key_store = {
-  keys : (string, key_store_key) Hashtbl.t;
+  keys : (string * key_store_key) list ref;
 }
 and key_store_key =
   | KeyString of string
   | KeyFileName of string
+  | KeyClevis
 
 external c_inspect_decrypt : Guestfs.t -> int64 -> (string * key_store_key) list -> unit = "guestfs_int_mllib_inspect_decrypt"
-external c_set_echo_keys : unit -> unit = "guestfs_int_mllib_set_echo_keys" "noalloc"
-external c_set_keys_from_stdin : unit -> unit = "guestfs_int_mllib_set_keys_from_stdin" "noalloc"
+external c_set_echo_keys : unit -> unit = "guestfs_int_mllib_set_echo_keys" [@@noalloc]
+external c_set_keys_from_stdin : unit -> unit = "guestfs_int_mllib_set_keys_from_stdin" [@@noalloc]
 external c_rfc3339_date_time_string : unit -> string = "guestfs_int_mllib_rfc3339_date_time_string"
 
 type machine_readable_fn = {
@@ -116,19 +123,59 @@ let message fs =
   in
   ksprintf display fs
 
+(* Wrap text. *)
+type wrap_break_t = WrapEOS | WrapSpace | WrapNL
+
+let rec wrap ?(chan = stdout) ?(indent = 0) str =
+  if Std_utils.wrap () || istty chan then
+    let len = String.length str in
+    _wrap chan indent 0 0 len str
+  else (
+    output_spaces chan indent;
+    output_string chan str
+  )
+
+and _wrap chan indent column i len str =
+  if i < len then (
+    let (j, break) = _wrap_find_next_break i len str in
+    let next_column =
+      if column + (j-i) >= 76 then (
+        output_char chan '\n';
+        output_spaces chan indent;
+        indent + (j-i) + 1
+      )
+      else column + (j-i) + 1 in
+    output chan (Bytes.of_string str) i (j-i);
+    match break with
+    | WrapEOS -> ()
+    | WrapSpace ->
+      output_char chan ' ';
+      _wrap chan indent next_column (j+1) len str
+    | WrapNL ->
+      output_char chan '\n';
+      output_spaces chan indent;
+      _wrap chan indent indent (j+1) len str
+  )
+
+and _wrap_find_next_break i len str =
+  if i >= len then (len, WrapEOS)
+  else if String.unsafe_get str i = ' ' then (i, WrapSpace)
+  else if String.unsafe_get str i = '\n' then (i, WrapNL)
+  else _wrap_find_next_break (i+1) len str
+
 (* Error messages etc. *)
 let error ?(exit_code = 1) fs =
   let display str =
     log_as_json "error" str;
     let chan = stderr in
     ansi_red ~chan ();
-    wrap ~chan (sprintf (f_"%s: error: %s") prog str);
+    wrap ~chan (sprintf (f_"%s: error: %s") !prog str);
     if not (verbose () && trace ()) then (
       prerr_newline ();
       prerr_newline ();
       wrap ~chan
            (sprintf (f_"If reporting bugs, run %s with debugging enabled and include the complete output:\n\n  %s -v -x [...]")
-                    prog prog);
+                    !prog !prog);
     );
     ansi_restore ~chan ();
     prerr_newline ();
@@ -141,7 +188,7 @@ let warning fs =
     log_as_json "warning" str;
     let chan = stdout in
     ansi_blue ~chan ();
-    wrap ~chan (sprintf (f_"%s: warning: %s") prog str);
+    wrap ~chan (sprintf (f_"%s: warning: %s") !prog str);
     ansi_restore ~chan ();
     print_newline ();
   in
@@ -152,7 +199,7 @@ let info fs =
     log_as_json "info" str;
     let chan = stdout in
     ansi_magenta ~chan ();
-    wrap ~chan (sprintf (f_"%s: %s") prog str);
+    wrap ~chan (sprintf (f_"%s: %s") !prog str);
     ansi_restore ~chan ();
     print_newline ();
   in
@@ -167,7 +214,8 @@ let debug fs =
  * (e.g. debug, tracing) already set.
  *)
 let open_guestfs ?identifier () =
-  let g = new Guestfs.guestfs () in
+  let g = new Guestfs.guestfs ~environment:false () in
+  g#parse_environment ();
   if trace () then g#set_trace true;
   if verbose () then g#set_verbose true;
   Option.may g#set_identifier identifier;
@@ -203,11 +251,11 @@ let run_main_and_handle_errors main =
  * the OCaml tools.
  *)
 let print_version_and_exit () =
-  printf "%s %s\n%!" prog Guestfs_config.package_version_full;
+  printf "%s %s\n%!" !prog Guestfs_config.package_version_full;
   exit 0
 
 let generated_by =
-  sprintf (f_"generated by %s %s") prog Guestfs_config.package_version_full
+  sprintf (f_"generated by %s %s") !prog Guestfs_config.package_version_full
 
 let virt_tools_data_dir =
   let dir = lazy (
@@ -293,12 +341,15 @@ let human_size i =
 type cmdline_options = {
   getopt : Getopt.t;
   ks : key_store;
+  debug_gc : bool ref;
 }
 
-let create_standard_options argspec ?anon_fun ?(key_opts = false) ?(machine_readable = false) usage_msg =
-  (** Install an exit hook to check gc consistency for --debug-gc *)
-  let set_debug_gc () =
-    at_exit (fun () -> Gc.compact()) in
+let create_standard_options argspec ?anon_fun ?(key_opts = false)
+      ?(machine_readable = false) ?(program_name = false) usage_msg =
+  (* Install an exit hook to check gc consistency for --debug-gc *)
+  let debug_gc = ref false in
+  let set_debug_gc () = at_exit Gc.compact; debug_gc := true in
+
   let parse_machine_readable = function
     | None ->
       machine_readable_output := Channel stdout
@@ -326,45 +377,63 @@ let create_standard_options argspec ?anon_fun ?(key_opts = false) ?(machine_read
       )
   in
   let ks = {
-    keys = Hashtbl.create 13;
+    keys = ref [];
   } in
-  let argspec = [
-    [ S 'V'; L"version" ], Getopt.Unit print_version_and_exit, s_"Display version and exit";
-    [ S 'v'; L"verbose" ], Getopt.Unit set_verbose,  s_"Enable libguestfs debugging messages";
-    [ S 'x' ],             Getopt.Unit set_trace,    s_"Enable tracing of libguestfs calls";
-    [ L"debug-gc" ],       Getopt.Unit set_debug_gc, Getopt.hidden_option_description;
-    [ S 'q'; L"quiet" ],   Getopt.Unit set_quiet,    s_"Don’t print progress messages";
-    [ L"color"; L"colors";
-      L"colour"; L"colours" ], Getopt.Unit set_colours, s_"Use ANSI colour sequences even if not tty";
-  ] @ argspec in
-  let argspec =
-    argspec @
-      (if key_opts then
-      let parse_key_selector arg =
-        let parts = String.nsplit ~max:3 ":" arg in
-        match parts with
-        | [ device; "key"; key ] ->
-          Hashtbl.replace ks.keys device (KeyString key)
-        | [ device; "file"; file ] ->
-          Hashtbl.replace ks.keys device (KeyFileName file)
-        | _ ->
-          error (f_"invalid selector string for --key: %s") arg
-      in
-      [
-        [ L"echo-keys" ],       Getopt.Unit c_set_echo_keys,       s_"Don’t turn off echo for passphrases";
-        [ L"keys-from-stdin" ], Getopt.Unit c_set_keys_from_stdin, s_"Read passphrases from stdin";
-        [ L"key" ], Getopt.String (s_"SELECTOR", parse_key_selector), s_"Specify a LUKS key";
-      ]
-      else []) @
-      (if machine_readable then
-      [
-        [ L"machine-readable" ], Getopt.OptString ("format", parse_machine_readable), s_"Make output machine readable";
-      ]
-      else []) in
+  let argspec = ref argspec in
+  let add_argspec = List.push_back argspec in
+
+  add_argspec ([ S 'V'; L"version" ], Getopt.Unit print_version_and_exit, s_"Display version and exit");
+  add_argspec ([ S 'v'; L"verbose" ], Getopt.Unit set_verbose,  s_"Enable libguestfs debugging messages");
+  add_argspec ([ S 'x' ],             Getopt.Unit set_trace,    s_"Enable tracing of libguestfs calls");
+  add_argspec ([ L"debug-gc" ],       Getopt.Unit set_debug_gc, Getopt.hidden_option_description);
+  add_argspec ([ S 'q'; L"quiet" ],   Getopt.Unit set_quiet,    s_"Don’t print progress messages");
+  add_argspec ([ L"color"; L"colors";
+                 L"colour"; L"colours" ], Getopt.Unit set_colours, s_"Use ANSI colour sequences even if not tty");
+  add_argspec ([ L"wrap" ],           Getopt.Unit set_wrap,     s_"Wrap log messages even if not tty");
+
+  if key_opts then (
+    let parse_key_selector arg =
+      let parts = String.nsplit ":" arg in
+      match parts with
+      | [] ->
+        error (f_"selector '%s': missing ID") arg
+      | [ _ ] ->
+        error (f_"selector '%s': missing TYPE") arg
+      | [ _; "key" ]
+      |  _ :: "key" :: _ :: _ :: _ ->
+        error (f_"selector '%s': missing KEY_STRING, or too many fields") arg
+      | [ device; "key"; key ] ->
+         List.push_back ks.keys (device, KeyString key)
+      | [ _; "file" ]
+      |  _ :: "file" :: _ :: _ :: _ ->
+        error (f_"selector '%s': missing FILENAME, or too many fields") arg
+      | [ device; "file"; file ] ->
+         List.push_back ks.keys (device, KeyFileName file)
+      |  _ :: "clevis" :: _ :: _ ->
+        error (f_"selector '%s': too many fields") arg
+      | [ device; "clevis" ] ->
+         List.push_back ks.keys (device, KeyClevis)
+      | _ ->
+         error (f_"selector '%s': invalid TYPE") arg
+    in
+
+    add_argspec ([ L"echo-keys" ],       Getopt.Unit c_set_echo_keys,       s_"Don’t turn off echo for passphrases");
+    add_argspec ([ L"keys-from-stdin" ], Getopt.Unit c_set_keys_from_stdin, s_"Read passphrases from stdin");
+    add_argspec ([ L"key" ], Getopt.String (s_"SELECTOR", parse_key_selector), s_"Specify a LUKS key");
+  );
+
+  if machine_readable then (
+    add_argspec ([ L"machine-readable" ], Getopt.OptString ("format", parse_machine_readable), s_"Make output machine readable");
+  );
+
+  if program_name then (
+    add_argspec ([ L"program-name" ], Getopt.Set_string ("prog", prog), s_"Set program name");
+  );
+
+  let argspec = !argspec in
+
   let getopt = Getopt.create argspec ?anon_fun usage_msg in
-  {
-    getopt; ks;
-  }
+  { getopt; ks; debug_gc }
 
 (* Run an external command, slurp up the output as a list of lines. *)
 let external_command ?(echo_cmd = true) cmd =
@@ -477,29 +546,6 @@ let uuidgen () =
   if String.length uuid < 10 then assert false; (* sanity check on uuidgen *)
   uuid
 
-(* Remove a temporary directory on exit. *)
-let rmdir_on_exit =
-  let dirs = ref [] in
-  let registered_handlers = ref false in
-
-  let rec rmdirs () =
-    List.iter (
-      fun dir ->
-        let cmd = sprintf "rm -rf %s" (Filename.quote dir) in
-        ignore (shell_command cmd)
-    ) !dirs
-  and register_handlers () =
-    (* Remove on exit. *)
-    at_exit rmdirs
-  in
-
-  fun dir ->
-    dirs := dir :: !dirs;
-    if not !registered_handlers then (
-      register_handlers ();
-      registered_handlers := true
-    )
-
 (* Using the libguestfs API, recursively remove only files from the
  * given directory.  Useful for cleaning /var/cache etc in sysprep
  * without removing the actual directory structure.  Also if 'dir' is
@@ -587,7 +633,7 @@ let debug_augeas_errors g =
 
       flush stderr
     with
-      Guestfs.Error msg -> eprintf "%s: augeas: %s (ignored)\n" prog msg
+      Guestfs.Error msg -> eprintf "%s: augeas: %s (ignored)\n" !prog msg
   )
 
 (* Detect type of a file. *)
@@ -650,21 +696,19 @@ let is_btrfs_subvolume g fs =
     if g#last_errno () = Guestfs.Errno.errno_EINVAL then false
     else raise exn
 
+let key_store_requires_network ks =
+  List.exists (function
+               | _, KeyClevis -> true
+               | _ -> false) !(ks.keys)
+
 let inspect_decrypt g ks =
-  (* Turn the keys in the key_store into a simpler struct, so it is possible
-   * to read it using the C API.
-   *)
-  let keys_as_list = Hashtbl.fold (
-    fun k v acc ->
-      (k, v) :: acc
-  ) ks.keys [] in
   (* Note we pass original 'g' even though it is not used by the
    * callee.  This is so that 'g' is kept as a root on the stack, and
    * so cannot be garbage collected while we are in the c_inspect_decrypt
    * function.
    *)
   c_inspect_decrypt g#ocaml_handle (Guestfs.c_pointer g#ocaml_handle)
-    keys_as_list
+    !(ks.keys)
 
 let with_timeout op timeout ?(sleep = 2) fn =
   let start_t = Unix.gettimeofday () in
