@@ -50,12 +50,50 @@ let convert (g : G.guestfs) source inspect keep_serial_console _ =
     match inspect.i_distro with
     | "fedora"
     | "rhel" | "centos" | "scientificlinux" | "redhat-based"
-    | "oraclelinux" -> `RHEL_family
+    | "oraclelinux" | "rocky" -> `RHEL_family
     | "altlinux" -> `ALT_family
     | "sles" | "suse-based" | "opensuse" -> `SUSE_family
     | "debian" | "ubuntu" | "linuxmint" | "kalilinux" -> `Debian_family
     | _ -> assert false in
 
+  (* map the OS family name to the qemu-guest-agent package name *)
+  let qga_pkg_of_family =
+    function
+    | `RHEL_family
+    | `ALT_family
+    | `SUSE_family
+    | `Debian_family -> Some "qemu-guest-agent"
+    | _ -> None
+  in
+
+  let qga_svc_start_cmd family distro major =
+    match family, distro, major with
+    | `RHEL_family, ( "rhel" | "centos" | "scientificlinux" | "redhat-based" |
+                      "oraclelinux" | "rocky" ), 6 ->
+      (* https://bugzilla.redhat.com/show_bug.cgi?id=2028764#c52 *)
+      Some "service qemu-ga start"
+
+    | `RHEL_family, _, _ ->
+      (* https://bugzilla.redhat.com/show_bug.cgi?id=2028764#c52 *)
+      Some "systemctl start qemu-guest-agent"
+
+    | `ALT_family, _, _ ->
+      (* https://bugzilla.redhat.com/show_bug.cgi?id=2028764#c45 *)
+      Some "systemctl start qemu-guest-agent"
+
+    | `SUSE_family, _, _ ->
+      (* https://bugzilla.redhat.com/show_bug.cgi?id=2028764#c51 *)
+      None
+
+    | `Debian_family, _, _ ->
+      (* https://bugzilla.redhat.com/show_bug.cgi?id=2028764#c42 *)
+      Some "service qemu-guest-agent start"
+
+    | _ ->
+      (* should never be called when "qga_pkg_of_family" returns None *)
+      assert false
+  in
+
   assert (inspect.i_package_format = "rpm" || inspect.i_package_format = "deb");
 
   (* Fail early if i_apps is empty.  Certain steps such as kernel
@@ -146,8 +184,8 @@ let convert (g : G.guestfs) source inspect keep_serial_console _ =
             *)
            (match inspect.i_distro, inspect.i_major_version with
             | "fedora", _ -> Q35
-            | ("rhel"|"centos"|"scientificlinux"|"redhat-based"|"oraclelinux"),
-              major ->
+            | ("rhel"|"centos"|"scientificlinux"|"redhat-based"|"oraclelinux"|
+               "rocky"), major ->
               if major <= 4 then I440FX else Q35
             | ("sles"|"suse-based"|"opensuse"), major ->
               if major < 10 then I440FX else Q35
@@ -539,14 +577,91 @@ let convert (g : G.guestfs) source inspect keep_serial_console _ =
 
   and install_linux_tools () =
     (* It is not fatal if we fail to install the QEMU guest agent. *)
-    let has_qemu_guest_agent =
-      List.exists (
-        fun { G.app2_name = name } ->
-          name = "qemu-guest-agent"
-      ) inspect.i_apps in
-    if not has_qemu_guest_agent then
-      (* FIXME -- install qemu-guest-agent here *)
-      ()
+    match qga_pkg_of_family family with
+    | None -> warning (f_"The name of the package that provides the QEMU Guest \
+                          Agent for this guest OS is unknown.  The guest agent \
+                          will not be installed.  Please consider reporting a \
+                          bug according to the BUGS section of the virt-v2v(1) \
+                          manual.")
+    | Some qga_pkg ->
+        let has_qemu_guest_agent =
+          List.exists (
+            fun { G.app2_name = name } ->
+              name = qga_pkg
+          ) inspect.i_apps in
+        if not has_qemu_guest_agent then
+          try
+            let inst_cmd = Guest_packages.install_command [qga_pkg]
+                             inspect.i_package_management in
+
+            (* Use only the portable filename character set in this. *)
+            let selinux_enforcing = "/root/virt-v2v-fb-selinux-enforcing"
+            and timeout = 30 in
+            let fbs =
+              Firstboot.add_firstboot_script g inspect.i_root
+            in
+            info (f_"The QEMU Guest Agent will be installed for this guest at \
+                     first boot.");
+
+            (* Wait for the network to come online in the guest (best effort).
+             *)
+            fbs "wait online"
+              (sprintf "#!/bin/sh\n\
+                        if conn=$(nmcli networking connectivity); then\n\
+                        \ \ tries=0\n\
+                        \ \ while\n\
+                        \ \ \ \ test $tries -lt %d &&\n\
+                        \ \ \ \ test full != \"$conn\"\n\
+                        \ \ do\n\
+                        \ \ \ \ sleep 1\n\
+                        \ \ \ \ tries=$((tries + 1))\n\
+                        \ \ \ \ conn=$(nmcli networking connectivity)\n\
+                        \ \ done\n\
+                        elif systemctl -q is-active systemd-networkd; then\n\
+                        \ \ /usr/lib/systemd/systemd-networkd-wait-online \\\n\
+                        \ \ \ \ -q --timeout=%d\n\
+                        fi\n" timeout timeout);
+
+            (* Disable SELinux temporarily around package installation. Refer to
+             * <https://bugzilla.redhat.com/show_bug.cgi?id=2028764#c7> and
+             * <https://bugzilla.redhat.com/show_bug.cgi?id=2028764#c8>.
+             *)
+            fbs "setenforce 0"
+              (sprintf "#!/bin/sh\n\
+                        rm -f %s\n\
+                        if command -v getenforce >/dev/null &&\n\
+                        \ \ test Enforcing = \"$(getenforce)\"\n\
+                        then\n\
+                        \ \ touch %s\n\
+                        \ \ setenforce 0\n\
+                        fi\n" selinux_enforcing selinux_enforcing);
+            fbs "install qga" inst_cmd;
+            fbs "setenforce restore"
+              (sprintf "#!/bin/sh\n\
+                        if test -f %s; then\n\
+                        \ \ setenforce 1\n\
+                        \ \ rm -f %s\n\
+                        fi\n" selinux_enforcing selinux_enforcing);
+
+            (* On all the distro families covered by "qga_pkg_of_family" and
+             * "qga_svc_start_cmd", the QEMU guest agent service is always
+             * enabled by package installation for *subsequent* boots. Package
+             * installation may or may not enable the service for the current
+             * (i.e., first) boot, however, so try that here manually.
+             *)
+            match qga_svc_start_cmd family inspect.i_distro inspect.i_major_version
+            with
+            | None -> ()
+            | Some start_cmd ->
+              fbs "start qga"
+                (sprintf "#!/bin/sh\n\
+                          %s\n" start_cmd)
+          with
+          | Guest_packages.Unknown_package_manager msg
+          | Guest_packages.Unimplemented_package_manager msg ->
+            warning (f_"The QEMU Guest Agent will not be installed.  The \
+                        install command for package ‘%s’ could not be created: \
+                        %s.") qga_pkg msg
 
   and configure_kernel () =
     (* Previously this function would try to install kernels, but we
@@ -616,8 +731,13 @@ let convert (g : G.guestfs) source inspect keep_serial_console _ =
     match kernel.ki_initrd with
     | None -> ()
     | Some initrd ->
-      (* Enable the basic virtio modules in the kernel. *)
-      (* Also forcibly include the "xts" module; see RHBZ#1658126. *)
+      (* Enable the basic virtio modules in the kernel.
+       *
+       * Also forcibly include the "xts" module; see RHBZ#1658126.
+       *
+       * Include the BOCHS DRM paravirt video driver; see RHBZ#2131123.  This
+       * driver is known under two names -- "bochs-drm" and "bochs".
+       *)
       let modules =
         let modules =
           (* The order of modules here is deliberately the same as the
@@ -628,7 +748,8 @@ let convert (g : G.guestfs) source inspect keep_serial_console _ =
            *)
           List.filter (fun m -> List.mem m kernel.ki_modules)
                       [ "virtio"; "virtio_ring"; "virtio_blk";
-                        "virtio_scsi"; "virtio_net"; "virtio_pci"; "xts" ] in
+                        "virtio_scsi"; "virtio_net"; "virtio_pci"; "xts";
+                        "bochs-drm"; "bochs" ] in
         if modules <> [] then modules
         else
           (* Fallback copied from old virt-v2v.  XXX Why not "ide"? *)
@@ -1108,6 +1229,7 @@ let convert (g : G.guestfs) source inspect keep_serial_console _ =
     (* Map device names for each entry. *)
     let rex_resume = PCRE.compile "^resume=(/dev/[-a-z\\d/_]+)(.*)$"
     and rex_device_cciss = PCRE.compile "^/dev/(cciss/c\\d+d\\d+)(?:p(\\d+))?$"
+    and rex_device_nvme = PCRE.compile "^/dev/(nvme\\d+n1)(?:p(\\d+))?$"
     and rex_device = PCRE.compile "^/dev/([a-z]+)(\\d*)?$" in
 
     let rec replace_if_device path value =
@@ -1130,6 +1252,11 @@ let convert (g : G.guestfs) source inspect keep_serial_console _ =
         and part = try PCRE.sub 2 with Not_found -> "" in
         "/dev/" ^ replace device ^ part
       )
+      else if PCRE.matches rex_device_nvme value then (
+        let device = PCRE.sub 1
+        and part = try PCRE.sub 2 with Not_found -> "" in
+        "/dev/" ^ replace device ^ part
+      )
       else if PCRE.matches rex_device value then (
         let device = PCRE.sub 1
         and part = try PCRE.sub 2 with Not_found -> "" in
@@ -1200,12 +1327,6 @@ let convert (g : G.guestfs) source inspect keep_serial_console _ =
         info (f_"Can't fix UEFI bootloader. VM may not boot.")
       in
 
-      let get_uefi_arch_suffix = function
-        | "x86_64" -> Some "X64"
-        | "i386" -> Some "X32"
-        | _ -> None
-      in
-
       match get_uefi_arch_suffix inspect.i_arch with
       | None -> cant_fix_uefi ()
       | Some suffix -> (
@@ -1305,11 +1426,11 @@ let convert (g : G.guestfs) source inspect keep_serial_console _ =
      * device names.  blkid will rebuild these on demand.
      *
      * Delete the LVM cache since it will contain references to the
-     * old devices (RHBZ#1164853).
+     * old devices (RHBZ#1164853, RHBZ#2112801).
      *)
     List.iter g#rm_f [
       "/etc/blkid/blkid.tab"; "/etc/blkid.tab";
-      "/etc/lvm/cache/.cache"
+      "/etc/lvm/cache/.cache"; "/etc/lvm/devices/system.devices"
     ];
   in
 
